from threading import Condition
import c4d
from cams_ids import *
from cams_utils_pref import *


DR_SUCCESS                           = 0
DR_ERROR_RENDER                      = 10000
DR_ERROR_STALE_DOC                   = 10001
DR_ERROR_IMAGE_ALLOC                 = 10002
DR_ERROR_UNKNOWN                     = 10003


## EXPECTS A CLONED, NON-INSERTED DOCUMENT, DOCUMENT WILL BE MODIFIED!!!
def RenderClonedDoc(doc, rdRender=None, flagsRenderAdditional=c4d.RENDERFLAGS_PREVIEWRENDER, thd=None, convertMultipassBitmap=False):
    if rdRender is None:
        rdRender = doc.GetActiveRenderData()

    else:
        doc.InsertRenderDataLast(rdRender)
        doc.SetActiveRenderData(rdRender)

    w = int(rdRender[c4d.RDATA_XRES])
    h = int(rdRender[c4d.RDATA_YRES])

    if rdRender[c4d.RDATA_ALPHACHANNEL]:
        bmpResult = c4d.bitmaps.MultipassBitmap(w, h, c4d.COLORMODE_ARGB)

    else:
        bmpResult = c4d.bitmaps.BaseBitmap()
        bmpResult.Init(w, h, 24)

    # RENDERFLAGS_NODOCUMENTCLONE: We are working with an already cloned document!
    flagsRender = c4d.RENDERFLAGS_EXTERNAL | c4d.RENDERFLAGS_NODOCUMENTCLONE | flagsRenderAdditional
    resultRender = c4d.documents.RenderDocument(doc, rdRender.GetData(), bmpResult, flagsRender, th=thd)

    if resultRender != c4d.RENDERRESULT_OK:
        PrintError('Deferred rendering: {0}'.format(resultRender))
        result = DR_ERROR_RENDER

    else:
        result = DR_SUCCESS

    if convertMultipassBitmap and result == DR_SUCCESS and bmpResult.IsMultipassBitmap():
        bmpTemp = c4d.bitmaps.BaseBitmap()
        bmpTemp.Init(w, h, 24)
        bmpTemp.AddChannel(True, False)

        bmpTempAlpha = bmpTemp.GetInternalChannel()

        bmpResultAlpha = bmpResult.GetInternalChannel()

        for y in range(h):
            for x in range(w):
                r, g, b = bmpResult.GetPixel(x, y)
                a = bmpResult.GetAlphaPixel(bmpResultAlpha, x, y)

                bmpTemp.SetPixel(x, y, r, g, b)
                bmpTemp.SetAlphaPixel(bmpTempAlpha, x, y, a)

        bmpResult = bmpTemp

    return bmpResult, result, resultRender


####################################################################################################
# Class RenderJob ##################################################################################
####################################################################################################
class RenderJob():
    _docRender = None   # Has to be a exclusively cloned document, which will be rendered
    _docOrig = None     # Reference to original document (source of clone), will be checked to be still alive and tunneled back to customer
    _idJob = None       # Arbitrary job ID, untouched and tunneled back to customer
    _rdCustom = None    # Optional custom RenderData for rendering
    _convertMultipassBitmap = False # Convert resulting MultipassBitmap to BaseBitmap
    _flags = 0          # Job flags, upper 16 bit untouched and tunneled back to customer
    _flagsRenderAdditional = 0  # Job flags, upper 16 bit untouched and tunneled back to customer

    _bmpResult = None
    _error = 0

    def __init__(self, docRender, docOrig, idJob, rdCustom=None, flags=0, flagsRenderAdditional=c4d.RENDERFLAGS_PREVIEWRENDER, convertMultipassBitmap=False):
        self._docOrig = docOrig
        self._docRender = docRender
        self._idJob = idJob
        self._rdCustom = rdCustom
        self._convertMultipassBitmap = convertMultipassBitmap
        self._flags = flags
        self._flagsRenderAdditional = flagsRenderAdditional
        self._bmpResult = None
        self._result = DR_ERROR_UNKNOWN
        self._resultRender = 0


    def HasError(self):
        return self._result != DR_SUCCESS


    def PrintError(self):
        if self._result == DR_SUCCESS:
            return

        if self._result == DR_ERROR_RENDER:
            PrintError('Render job failed: Render Error: {0}'.format(self._resultRender))

        elif self._result == DR_ERROR_STALE_DOC:
            PrintError('Render job failed: Stale document')

        elif self._result == DR_ERROR_IMAGE_ALLOC:
            PrintError('Render job failed: Bitmap allocation failed')

        elif self._result == DR_ERROR_UNKNOWN:
            PrintError('Render job failed: UNKNOWN ERROR - report bug')

        else:
            PrintError('Render job failed: UNEXPECTED ERROR - report bug')


####################################################################################################
# Class RenderThreasd ##############################################################################
####################################################################################################
class RenderThread(c4d.threading.C4DThread):
    _parent = None
    _run = True

    def Main(self):
        while self._run:
            job = self._parent.WaitForJob()

            if self.TestBreak() or not self._run:
                break

            if job is None:
                continue

            docClone = job._docRender
            doc = job._docOrig

            if doc is None or not doc.IsAlive() or docClone is None or not docClone.IsAlive():
                job._bmpResult = None
                job._error = DR_ERROR_STALE_DOC

                self._parent.QueueResult(job)

                continue

            job._bmpResult, job._result, job._resultRender = RenderClonedDoc(docClone, job._rdCustom, job._flagsRenderAdditional, thd=self.Get(), convertMultipassBitmap=job._convertMultipassBitmap)

            self._parent.QueueResult(job)


####################################################################################################
# Class DeferredRendering ##########################################################################
####################################################################################################
class DeferredRendering():
    _idCoreMessage = None
    _subIdCoreMessageRenderFinished = None
    _subIdCoreMessageThreadRestarted = None

    _mutexQueues = None

    _queueJobs = []
    _queueResults = []

    _thdRender = None

    _isBusy = False

    def __init__(self, idCoreMessage, subIdCoreMessageRenderFinished, subIdCoreMessageThreadRestarted, subIdCoreMessageDeferredRenderIdle=None, subIdCoreMessageDeferredRenderBusy=None):
        self._idCoreMessage = idCoreMessage
        self._subIdCoreMessageRenderFinished = subIdCoreMessageRenderFinished
        self._subIdCoreMessageThreadRestarted = subIdCoreMessageThreadRestarted
        self._subIdCoreMessageDeferredRenderIdle = subIdCoreMessageDeferredRenderIdle
        self._subIdCoreMessageDeferredRenderBusy = subIdCoreMessageDeferredRenderBusy

        self._mutexQueues = Condition()

        self._queueJobs = []
        self._queueResults = []

        self._thdRender = RenderThread()

        self._thdRender._parent = self
        self._thdRender._run = True

        self._thdRender.Start()

        self._isBusy = False


    def AddJob(self, doc, idJob, docClone=None, rdCustom=None, flags=0, flagsRenderAdditional=c4d.RENDERFLAGS_PREVIEWRENDER, convertMultipassBitmap=False):
        if not ReadPrefValue(ID_PREF_THUMB_ENABLED):
            return

        if docClone is None:
            docClone = doc.GetClone()

        self._mutexQueues.acquire()

        try:
            self._queueJobs.append(RenderJob(docClone, doc, idJob, rdCustom, flags, flagsRenderAdditional, convertMultipassBitmap))
            self._mutexQueues.notify()

            if self._subIdCoreMessageDeferredRenderBusy is not None:
                c4d.SpecialEventAdd(self._idCoreMessage, self._subIdCoreMessageDeferredRenderBusy)
            self._isBusy = True

        except Exception as e:
            PrintError(e)

        self._mutexQueues.release()


    def WaitForJob(self):
        job = None

        self._mutexQueues.acquire()

        try:
            if len(self._queueJobs) <= 0:
                if self._subIdCoreMessageDeferredRenderIdle is not None:
                    c4d.SpecialEventAdd(self._idCoreMessage, self._subIdCoreMessageDeferredRenderIdle)
                self._isBusy = False

                self._mutexQueues.wait()

            if len(self._queueJobs) > 0:
                job = self._queueJobs.pop(0)

        except Exception as e:
            PrintError(e)

        self._mutexQueues.release()

        return job


    def QueueResult(self, jobResult):
        self._mutexQueues.acquire()

        try:
            self._queueResults.append(jobResult)

        except Exception as e:
            PrintError(e)

        self._mutexQueues.release()

        c4d.SpecialEventAdd(self._idCoreMessage, p1=self._subIdCoreMessageRenderFinished)


    def GetResult(self):
        jobResult = None

        self._mutexQueues.acquire()

        try:
            if len(self._queueResults) > 0:
                jobResult = self._queueResults.pop(0)

        except Exception as e:
            PrintError(e)

        self._mutexQueues.release()

        return jobResult


    def Reset(self):
        self._mutexQueues.acquire()

        self._queueJobs = []
        self._queueResults = []

        self._mutexQueues.release()


    def StopRenderThread(self):
        self._thdRender._run = False

        self._thdRender.End(False)

        self._mutexQueues.acquire()

        self._mutexQueues.notify()

        self._mutexQueues.release()

        self._thdRender.End(True)


    def RestartRenderThread(self):
        if not self._thdRender._run:
            self._thdRender._run = True

            self._thdRender.Start()

            c4d.SpecialEventAdd(self._idCoreMessage, p1=self._subIdCoreMessageThreadRestarted)


    def IsBusy(self):
        return self._isBusy
