import math
import c4d
from cams_utils_image import LoadImage

PI2 = 2.0 * math.pi
PI_05 = 0.5 * math.pi


class TagPreviewCamera(c4d.plugins.TagData):

    _bmpIconTurntable = None
    _bmpIconLookAround = None


    def Init(self, node):
        self.InitAttr(node, int,   c4d.CAMS_PREVIEW_CAMERA_MODE)
        self.InitAttr(node, int,   c4d.CAMS_PREVIEW_CAMERA_DURATION)
        self.InitAttr(node, int,   c4d.CAMS_PREVIEW_CAMERA_DIR)
        self.InitAttr(node, float, c4d.CAMS_PREVIEW_CAMERA_ANGLE_START)
        self.InitAttr(node, float, c4d.CAMS_PREVIEW_CAMERA_HEIGHT)
        self.InitAttr(node, float, c4d.CAMS_PREVIEW_CAMERA_RADIUS)
        self.InitAttr(node, float, c4d.CAMS_PREVIEW_CAMERA_TARGET_HEIGHT)

        doc = c4d.documents.GetActiveDocument()
        timeMax = doc.GetMaxTime().GetFrame(doc.GetFps())

        node[c4d.CAMS_PREVIEW_CAMERA_MODE] = c4d.CAMS_PREVIEW_CAMERA_MODE_OUTSIDE
        node[c4d.CAMS_PREVIEW_CAMERA_DURATION] = timeMax
        node[c4d.CAMS_PREVIEW_CAMERA_DIR] = c4d.CAMS_PREVIEW_CAMERA_DIR_CCW
        node[c4d.CAMS_PREVIEW_CAMERA_ANGLE_START] = 0.0
        node[c4d.CAMS_PREVIEW_CAMERA_HEIGHT] = 0.0
        node[c4d.CAMS_PREVIEW_CAMERA_RADIUS] = 500.0
        node[c4d.CAMS_PREVIEW_CAMERA_TARGET_HEIGHT] = 0.0

        return True


    def GetDEnabling(self, node, id, t_data, flags, itemdesc):
        doc = node.GetDocument()

        bcParams = node.GetDataInstance()

        mode = bcParams.GetInt32(c4d.CAMS_PREVIEW_CAMERA_MODE)
        objTarget = bcParams.GetLink(c4d.CAMS_PREVIEW_CAMERA_TARGET, doc)
        objHeightHelper = bcParams.GetLink(c4d.CAMS_PREVIEW_CAMERA_HEIGHT_HELPER, doc)
        objParent = node.GetObject().GetUp()

        isModeOutside = mode == c4d.CAMS_PREVIEW_CAMERA_MODE_OUTSIDE
        hasTarget = objTarget is not None
        hasHeightHelper = objHeightHelper is not None
        hasParent = objParent is not None

        if id[0].id == c4d.CAMS_PREVIEW_CAMERA_MODE:
            return True

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_DURATION:
            return not isModeOutside or hasTarget

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_DIR:
            return not isModeOutside or hasTarget

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_ANGLE_START:
            return (isModeOutside and hasTarget) or not isModeOutside

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_HEIGHT:
            return (isModeOutside and hasTarget) or (not isModeOutside and hasParent)

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_RADIUS:
            return (isModeOutside and hasTarget) or (not isModeOutside and hasParent)

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_TARGET:
            return isModeOutside

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_TARGET_HEIGHT:
            return (isModeOutside and hasTarget and not hasHeightHelper) or not isModeOutside

        elif id[0].id == c4d.CAMS_PREVIEW_CAMERA_HEIGHT_HELPER:
            return isModeOutside

        else:
            return hasTarget

        return True


    def ExecuteOutsideIn(self, tag, doc, op):
        bcParams = tag.GetDataInstance()

        objTarget = bcParams.GetObjectLink(c4d.CAMS_PREVIEW_CAMERA_TARGET, doc)
        if objTarget is None:
            return c4d.EXECUTIONRESULT_OK

        mgTarget = objTarget.GetMg()
        vPosTarget = mgTarget.off

        numFrames = bcParams[c4d.CAMS_PREVIEW_CAMERA_DURATION]
        idxFrameCurrent = doc.GetTime().GetFrame(doc.GetFps())

        objHeightHelper = bcParams.GetLink(c4d.CAMS_PREVIEW_CAMERA_HEIGHT_HELPER, doc)
        hasHeightHelper = objHeightHelper is not None

        if hasHeightHelper:
            mgHeightHelper = objHeightHelper.GetMg()
            bcParams[c4d.CAMS_PREVIEW_CAMERA_TARGET_HEIGHT] = mgHeightHelper.off.y - vPosTarget.y

        direction = bcParams[c4d.CAMS_PREVIEW_CAMERA_DIR]
        angleStart = bcParams[c4d.CAMS_PREVIEW_CAMERA_ANGLE_START]
        radius = max(0.0001, bcParams[c4d.CAMS_PREVIEW_CAMERA_RADIUS])
        height = bcParams[c4d.CAMS_PREVIEW_CAMERA_HEIGHT]
        targetHeight = bcParams[c4d.CAMS_PREVIEW_CAMERA_TARGET_HEIGHT]

        hDiff = targetHeight - height

        x = math.sqrt(hDiff * hDiff + radius * radius)

        pitch = math.asin(hDiff / x)

        progress = float(idxFrameCurrent % numFrames) / float(numFrames)

        h = angleStart + (progress * PI2)

        if direction == c4d.CAMS_PREVIEW_CAMERA_DIR_CW:
            h = -h

        posX = math.cos(h) * radius
        posZ = math.sin(h) * radius

        vRot = c4d.Vector(h + PI_05, pitch, 0.0)
        vPos = c4d.Vector(vPosTarget.x + posX, vPosTarget.y + height, vPosTarget.z + posZ)

        mgCamera = c4d.utils.HPBToMatrix(vRot)
        mgCamera.off = vPos

        op.SetMg(mgCamera)

        return c4d.EXECUTIONRESULT_OK


    def ExecuteInsideOut(self, tag, doc, op):
        bcParams = tag.GetDataInstance()

        numFrames = bcParams[c4d.CAMS_PREVIEW_CAMERA_DURATION]
        idxFrameCurrent = doc.GetTime().GetFrame(doc.GetFps())

        direction = bcParams[c4d.CAMS_PREVIEW_CAMERA_DIR]
        angleStart = bcParams[c4d.CAMS_PREVIEW_CAMERA_ANGLE_START]
        targetHeight = bcParams[c4d.CAMS_PREVIEW_CAMERA_TARGET_HEIGHT]

        objParent = op.GetUp()

        if objParent is None:
            radius = 0.0
            height = 0.0
            vPos = op.GetMg().off

        else:
            radius = bcParams[c4d.CAMS_PREVIEW_CAMERA_RADIUS]
            height = bcParams[c4d.CAMS_PREVIEW_CAMERA_HEIGHT]
            vPos = objParent.GetMg().off

        x = math.sqrt(targetHeight * targetHeight + 1000000.0)

        pitch = math.asin(targetHeight / x)

        progress = float(idxFrameCurrent % numFrames) / float(numFrames)

        h = angleStart + (progress * PI2)

        if direction == c4d.CAMS_PREVIEW_CAMERA_DIR_CW:
            h = -h

        vRot = c4d.Vector(h - PI_05, pitch, 0.0)

        if radius != 0.0:
            vPos.x = vPos.x + math.cos(h) * radius
            vPos.y = vPos.y + height
            vPos.z = vPos.z + math.sin(h) * radius

        mgCamera = c4d.utils.HPBToMatrix(vRot)
        mgCamera.off = vPos

        op.SetMg(mgCamera)

        return c4d.EXECUTIONRESULT_OK


    def Execute(self, tag, doc, op, bt, priority, flags):
        if tag[c4d.CAMS_PREVIEW_CAMERA_MODE] == c4d.CAMS_PREVIEW_CAMERA_MODE_OUTSIDE:
            return self.ExecuteOutsideIn(tag, doc, op)

        else:
            return self.ExecuteInsideOut(tag, doc, op)


    def Message(self, node, type, data):
        if type == c4d.MSG_GETCUSTOMICON:
            if self._bmpIconTurntable is None:
                self._bmpIconTurntable = LoadImage('cam_Turnaround', guiPixelRatio=2.0)

            if self._bmpIconLookAround is None:
                self._bmpIconLookAround = LoadImage('cam_Lookaround', guiPixelRatio=2.0)

            doc = node.GetDocument()

            bcParams = node.GetDataInstance()
            isModeOutside = bcParams[c4d.CAMS_PREVIEW_CAMERA_MODE] == c4d.CAMS_PREVIEW_CAMERA_MODE_OUTSIDE

            if isModeOutside:
                data['bmp'] = self._bmpIconTurntable

            else:
                data['bmp'] = self._bmpIconLookAround

            w, h = data['bmp'].GetSize()

            data['x'] = 0
            data['y'] = 0
            data['w'] = w
            data['h'] = h

            objTarget = bcParams.GetObjectLink(c4d.CAMS_PREVIEW_CAMERA_TARGET, doc)

            if c4d.GetC4DVersion() // 1000 > 20:
                if isModeOutside and objTarget is None:
                    flags = c4d.ICONDATAFLAGS_DISABLED

                else:
                    flags = c4d.ICONDATAFLAGS_NONE

            else:
                flags = 0

            data['flags'] = flags

            data['filled'] = True

        #elif type == c4d.MSG_MENUPREPARE:

        return True
